@@ -51,6 +51,8 @@
 
 QT_BEGIN_NAMESPACE
 
+Q_LOGGING_CATEGORY(lcDimmer, "qt.quick.controls.popup.dimmer")
+
 /*!
     \qmltype Popup
     \inherits QtObject
@@ -215,6 +217,19 @@ QT_BEGIN_NAMESPACE
     To ensure that the popup is positioned within the bounds of the enclosing
     window, the \l margins property can be set to a non-negative value.
 
+    \section1 Popup Transitions
+
+    Since Qt 5.15.3 the following properties are restored to their original values from before
+    the enter transition after the exit transition is completed.
+
+    \list
+    \li \l opacity
+    \li \l scale
+    \endlist
+
+    This allows the built-in styles to animate on these properties without losing any explicitly
+    defined value.
+
     \sa {Popup Controls}, {Customizing Popup}, ApplicationWindow
 */
 
@@ -441,6 +456,9 @@ bool QQuickPopupPrivate::prepareEnterTransition()
         popupItem->setVisible(true);
         getPositioner()->setParentItem(parentItem);
         emit q->visibleChanged();
+
+        if (focus)
+            popupItem->setFocus(true);
     }
     return true;
 }
@@ -451,6 +469,11 @@ bool QQuickPopupPrivate::prepareExitTransition()
     if (transitionState == ExitTransition && transitionManager.isRunning())
         return false;
 
+    // We need to cache the original scale and opacity values so we can reset it after
+    // the exit transition is done so they have the original values again
+    prevScale = popupItem->scale();
+    prevOpacity = popupItem->opacity();
+
     if (transitionState != ExitTransition) {
         // The setFocus(false) call below removes any active focus before we're
         // able to check it in finalizeExitTransition.
@@ -469,8 +492,6 @@ bool QQuickPopupPrivate::prepareExitTransition()
 void QQuickPopupPrivate::finalizeEnterTransition()
 {
     Q_Q(QQuickPopup);
-    if (focus)
-        popupItem->setFocus(true);
     transitionState = NoTransition;
     getPositioner()->reposition();
     emit q->openedChanged();
@@ -481,8 +502,10 @@ void QQuickPopupPrivate::finalizeExitTransition()
 {
     Q_Q(QQuickPopup);
     getPositioner()->setParentItem(nullptr);
-    popupItem->setParentItem(nullptr);
-    popupItem->setVisible(false);
+    if (popupItem) {
+        popupItem->setParentItem(nullptr);
+        popupItem->setVisible(false);
+    }
     destroyOverlay();
 
     if (hadActiveFocusBeforeExitTransition && window) {
@@ -491,13 +514,14 @@ void QQuickPopupPrivate::finalizeExitTransition()
         if (QQuickOverlay *overlay = QQuickOverlay::overlay(window)) {
             const auto stackingOrderPopups = QQuickOverlayPrivate::get(overlay)->stackingOrderPopups();
             for (auto popup : stackingOrderPopups) {
-                if (QQuickPopupPrivate::get(popup)->transitionState != ExitTransition) {
+                if (QQuickPopupPrivate::get(popup)->transitionState != ExitTransition
+                         && popup->hasFocus()) {
                     nextFocusPopup = popup;
                     break;
                 }
             }
         }
-        if (nextFocusPopup && nextFocusPopup->hasFocus()) {
+        if (nextFocusPopup) {
             nextFocusPopup->forceActiveFocus();
         } else {
             QQuickApplicationWindow *applicationWindow = qobject_cast<QQuickApplicationWindow*>(window);
@@ -513,6 +537,10 @@ void QQuickPopupPrivate::finalizeExitTransition()
     hadActiveFocusBeforeExitTransition = false;
     emit q->visibleChanged();
     emit q->closed();
+    if (popupItem) {
+        popupItem->setScale(prevScale);
+        popupItem->setOpacity(prevOpacity);
+    }
 }
 
 QMarginsF QQuickPopupPrivate::getMargins() const
@@ -575,7 +603,6 @@ void QQuickPopupPrivate::setBottomMargin(qreal value, bool reset)
 
 /*!
     \since QtQuick.Controls 2.5 (Qt 5.12)
-    \qmlpropertygroup QtQuick.Controls::Popup::anchors
     \qmlproperty Object QtQuick.Controls::Popup::anchors.centerIn
 
     Anchors provide a way to position an item by specifying its
@@ -703,6 +730,8 @@ static QQuickItem *createDimmer(QQmlComponent *component, QQuickPopup *popup, QQ
         if (component)
             component->completeCreate();
     }
+    qCDebug(lcDimmer) << "finished creating dimmer from component" << component
+        << "for popup" << popup << "with parent" << parent << "- item is:" << item;
     return item;
 }
 
@@ -729,6 +758,7 @@ void QQuickPopupPrivate::createOverlay()
 void QQuickPopupPrivate::destroyOverlay()
 {
     if (dimmer) {
+        qCDebug(lcDimmer) << "destroying dimmer" << dimmer;
         dimmer->setParentItem(nullptr);
         dimmer->deleteLater();
         dimmer = nullptr;
@@ -827,6 +857,12 @@ QQuickPopup::~QQuickPopup()
     d->popupItem = nullptr;
     delete d->positioner;
     d->positioner = nullptr;
+
+    // If the popup is destroyed before the exit transition finishes,
+    // the necessary cleanup (removing modal dimmers that block mouse events,
+    // emitting closed signal, etc.) won't happen. That's why we do it manually here.
+    if (d->transitionState == QQuickPopupPrivate::ExitTransition && d->transitionManager.isRunning())
+        d->finalizeExitTransition();
 }
 
 /*!
